/*
	Public Domain Code
	Coded by The IceBreaker
*/

#include "model3ds.h"

/*
	This function will fill the given model structure
	with the contents of the file specified by FileName.
*/
nboolean model_Load(Model *model, const char *FileName)
{
	int				i;
	FILE			*pFile;
	unsigned short	usChunkID;
	unsigned int	uiChunkLength;
	unsigned char	uChar;
	unsigned short	usNum;
	unsigned short	usFaceFlags;
	unsigned int	uiVersion;
	unsigned short	uiBytesRead;
	unsigned short	a,b,c;
	float 			x,y,z,u,v;

	/* Set the number of meshes inside the model to 0 */
	model->NumMeshes = 0;
	
	if ((pFile=fopen(FileName, "rb"))==NULL) return false;

	/* Get the file size */
	long int iOldPos      = fseek(pFile,0L,SEEK_END);
	long int iFileLength  = ftell(pFile);
	fseek(pFile,iOldPos,SEEK_SET);

	/* Loop until current position is smaller than the file size */
	while (ftell(pFile)<iFileLength)
	{
		/* 
			Read the ChunkID and ChunkLength.
			
			We are catching the number of bytes readed
			by these are not important in our case, we
			could also write the first line as:
			
			fread(&usChunkID,sizeof(unsigned short),1,pFile);
		*/
		uiBytesRead=fread(&usChunkID,1,sizeof(unsigned short),pFile);
		uiBytesRead+=fread(&uiChunkLength,1,sizeof(unsigned int),pFile);

		switch (usChunkID)
		{
			/* 
				We end up here only once, but we don't need to 
				process anything anyway.
			*/
			case MAIN_CHUNK:
			break;

			/*
				It is good to check for the version number.
				If it is bigger than 3 we just display a
				warning messagebox, but we continue as usual.
			*/
			case VERSION:
			{
				/* version checking */
				uiBytesRead+=fread(&uiVersion,1,sizeof(unsigned int),pFile);
				if (uiVersion > 3)
				{
					MessageBox(0,"File version is >3, the file may not load correctly!","Warning",MB_ICONWARNING);					
				}
			}
			break;

			/* 
				We end up here only once, but we don't need to 
				process anything anyway.
			*/
			case EDITOR_CHUNK:
			break;

			/*
				The OBJECT_BLOCK chunk is a very important one,
				because here we alloc and resize the dynamic
				array of meshes, and also we read the actual
				mesh name from the file.
			*/
			case OBJECT_BLOCK:
			{
				if (model->NumMeshes == 0)
				{
					model->NumMeshes++;
					model->Meshes = (Mesh *) malloc(model->NumMeshes*sizeof(Mesh));
				}
				else
				{
					model->NumMeshes++;
					model->Meshes = (Mesh *) realloc(model->Meshes,model->NumMeshes*sizeof(Mesh));
				}

				i=0;
				do
				{
					uiBytesRead+=fread(&uChar,1,sizeof(unsigned char),pFile);
					model->Meshes[model->NumMeshes-1].MeshName[i]=uChar;
					i++;
				}while(uChar != '\0' && i<64);
			}
			break;

			/*
				We endup here multiple times, but we don't give
				a damn' ... hehe
			*/
			case TRIANGULAR_MESH:
			break;

			/*
			This chunk contains the actual "meat" of the file, the VERTICES.
			*/
			case VERTICES_LIST:
			{
				/* We already allocated meshes?!!! */
				if (model->NumMeshes > 0)
				{
					/* Get get the number of vertices for the current mesh */	
					uiBytesRead+=fread(&usNum, 1, sizeof(unsigned short), pFile);

					/* Allocate the dynamic vertex array */
					model->Meshes[model->NumMeshes-1].Vertices  = (Vertex *) malloc(usNum*sizeof(Vertex));	
					memset(model->Meshes[model->NumMeshes-1].Vertices,0,usNum*sizeof(Vertex));

					/* 
						Read the vertices as three floating point values, 
						and store them inside the allocated array.
					*/
					for (i=0; i<usNum; i++)
					{
						uiBytesRead+=fread(&x,1, sizeof(float),pFile);
						uiBytesRead+=fread(&y,1, sizeof(float),pFile);
						uiBytesRead+=fread(&z,1, sizeof(float),pFile);

						model->Meshes[model->NumMeshes-1].Vertices[i].x = x;
						model->Meshes[model->NumMeshes-1].Vertices[i].y = y;
						model->Meshes[model->NumMeshes-1].Vertices[i].z = z;
					}

					/*
						Finally store the number of allocated vertices.
					*/
					model->Meshes[model->NumMeshes-1].NumVertices = usNum;
				}
			}
			break;

			/*
				This chunk contains the face indices.
			*/
			case FACES_DESCRIPTION:
			{
				/* We already allocated meshes?!!! */
				if (model->NumMeshes > 0)
				{
					/* Get get the number of faces (triangles) for the current mesh */		
					uiBytesRead+=fread(&usNum,1,sizeof(unsigned short),pFile);

					/* Allocate the dynamic vertex array */
					model->Meshes[model->NumMeshes-1].Faces = (Face *) malloc(usNum*sizeof(Face));
					memset(model->Meshes[model->NumMeshes-1].Faces,0,usNum*sizeof(Face));

					/*
						Read the face indices as 3 unsigned shorts + 1 unsigned short used as
						a flag, but which is not important to us, but we have to read it, 
						because we like the KISS principle.

						Then obviously we store only the three indices inside the 
						allocated array.
					*/
					
					for (i=0;i<usNum; i++)
					{
						uiBytesRead+=fread(&a,1,sizeof(unsigned short),pFile);
						uiBytesRead+=fread(&b,1,sizeof(unsigned short),pFile);
						uiBytesRead+=fread(&c,1,sizeof(unsigned short),pFile);
						uiBytesRead+=fread(&usFaceFlags,1,sizeof(unsigned short),pFile);

						model->Meshes[model->NumMeshes-1].Faces[i].a = a;
						model->Meshes[model->NumMeshes-1].Faces[i].b = b;
						model->Meshes[model->NumMeshes-1].Faces[i].c = c;
					}

					/*
						Finally store the number of allocated faces.
					*/	
					model->Meshes[model->NumMeshes-1].NumFaces = usNum;
				}
			}
			break;

			/*
				This chunk contains the actual texture coordinates for
				the current mesh.

				The number of texture coordinates is equal to the number
				of vertices inside the mesh.
			*/
			case MAPPING_COORDINATES_LIST:
			{
				/* We already allocated meshes?!!! */
				if (model->NumMeshes > 0)
				{
					/* Read the number of texture coordinates ( == NumVertices ) */
					uiBytesRead+=fread(&usNum,1,sizeof(unsigned short), pFile);

					/* 
						Read the coordinates as two floating point values
						and store them inside the previously allocated
						dynamic vertex array.
					*/
					for (i=0; i<usNum; i++)
					{
						uiBytesRead+=fread(&u,1,sizeof(float),pFile);
						uiBytesRead+=fread(&v,1,sizeof(float),pFile);

						model->Meshes[model->NumMeshes-1].Vertices[i].u = u;
						model->Meshes[model->NumMeshes-1].Vertices[i].v = v;
					}
				}
			}
			break;

			default:
				/* 
					We deal like this with unknow chunks ... Grrr

					If we end up here uiBytesRead == 6, so we could	
					write uiChunkLengh-6 instead.

					How do I know that it will be 6?
					It is simple.At the beginning of the loop we read
					6 bytes in total. One unsigned short which is 2 bytes
					and one unsigned int which is 4 bytes.
				*/
				fseek(pFile, uiChunkLength-uiBytesRead, SEEK_CUR);
		}
	}

	fclose(pFile);

	/* 
		Play a bit with the position
		and rotation of the whole
		model.
	*/

	model->position[0] = 0.0f;
	model->position[1] = 0.0f;
	model->position[2] = -200.0f;

	model->rotation[0] = -90.0f;
	model->rotation[1] = 0.0f;
	model->rotation[2] = 0.0f;

	/*
		Assign the name of the first mesh
		as the global mesh name.
	*/
	if (model->NumMeshes > 0)
		strcpy(model->ModelName, model->Meshes[0].MeshName);

	return true; /* The model was loaded! */
}

/*
	This function will free the dynamic
	array allocated inside the given
	model structure.
*/
void model_Free(Model *model)
{
	int i;
	
	/* No mesh found in the model */
	if(!model->Meshes)
		return;
	
	for(i=0;i<model->NumMeshes;i++)
	{
			free(model->Meshes[i].Faces);
			free(model->Meshes[i].Vertices);
	}
	
	free(model->Meshes);
}

/*	
	This function will render the model inside
	the given model structure.
*/
void model_Render(Model *model)
{
	int i,index,index2,k;

	glPushMatrix();

	/*
		Positionate and rotate.
	*/
	glTranslatef(model->position[0],model->position[1],model->position[2]);
	glRotatef(model->rotation[0],1.0f,0.0f,0.0f);
	glRotatef(model->rotation[1],0.0f,1.0f,0.0f);
	glRotatef(model->rotation[2],0.0f,0.0f,1.0f);

	glBegin(GL_TRIANGLES);

		/* Loop through the meshes */
		for (i=0;i<model->NumMeshes;i++)
		{
			/* 
				Loop through the faces of the current mesh.
				
				You could also bind textures the each mesh here.
			*/
			for (index=0;index<model->Meshes[i].NumFaces;index++)
			{
					/*
						If you would want just face normals, then
						you should add only 1 glNormal3f()
						after the next line.
						
						For vertex normals you should add one
						glNormal3f() before each glTexCoord2f().
					*/
					
					index2 = model->Meshes[i].Faces[index].a;

					glTexCoord2f(
									model->Meshes[i].Vertices[index2].u,
									model->Meshes[i].Vertices[index2].v
								);
					glVertex3f	(
									model->Meshes[i].Vertices[index2].x,
									model->Meshes[i].Vertices[index2].y,
									model->Meshes[i].Vertices[index2].z
								);
					
					index2 = model->Meshes[i].Faces[index].b;

					glTexCoord2f(
									model->Meshes[i].Vertices[index2].u,
									model->Meshes[i].Vertices[index2].v
								);
					glVertex3f	(
									model->Meshes[i].Vertices[index2].x,
									model->Meshes[i].Vertices[index2].y,
									model->Meshes[i].Vertices[index2].z
								);
					index2 = model->Meshes[i].Faces[index].c;

					glTexCoord2f(
									model->Meshes[i].Vertices[index2].u,
									model->Meshes[i].Vertices[index2].v
								);
					glVertex3f	(
									model->Meshes[i].Vertices[index2].x,
									model->Meshes[i].Vertices[index2].y,
									model->Meshes[i].Vertices[index2].z
								);
			}
		}
		
	glEnd();
	glPopMatrix();
}

/*
	This function will increment the rotation vector's
	elements with the given x,y or z.
*/
void model_Rotate(Model *model,float x, float y, float z)
{
	model->rotation[0] += x;	
	model->rotation[1] += y;	
	model->rotation[2] += z;	
}
